#!/usr/bin/python

def Translate(line):
    return line.replace("&","&amp;");


class Function:
    """ Create a function desc, using Function(name,rtype)
    Add arguments in the <obj>.Args.append() """
    def __init__(self,name,rtype,ismethod=False):
        self.Name=name;
        self.Rtype=rtype;
        self.Args=[] #list of dicts
        self.IsMethod=ismethod
        self.IsVirtual=False;
        self.IsStatic=not self.IsMethod;
    
    def emit_xml(self):
        endstr=""
        if (self.IsMethod):
            print "<method virtual=\"%d\" name=\"%s\" rtype=\"%s\" >"%(self.IsVirtual,Translate(self.Name),Translate(self.Rtype))
            endstr="</method>"
        elif(self.IsStatic):
            print "<function name=\"%s\" rtype=\"%s\" >"%(Translate(self.Name),Translate(self.Rtype))
            endstr="</function>"

        #TODO: Add support for Default arguments.
        for i in self.Args:
            key,val=i.items()[0]
            print "<arg type=\"%s\" name=\"%s\" />"%(Translate(key),Translate(val))
                    
        print endstr
        

    def emit_octave(self,ClassName):
        print "DEFUN_DLD(%s_%s,args,,\"%s\") {" %(ClassName,self.Name,ClassName+"_"+self.Name)
                
        if (self.IsMethod):
            classvar="_"+ClassName
            empty=0;
            print "%s %s;"%(ClassName,classvar);
            for i in self.Args:
                key,val=i.items()[0]
                if key=='':
                    empty=empty+1;
                    continue
                print "%s %s;"%(key,val);

            print "if (args.length() != %d ) {"%(len(self.Args)+1-empty)
            print "%s"%(self.__repr__())
            print "\t}"
        elif(self.IsStatic):
            
            

        #TODO: Add support for Default arguments.
        for i in self.Args:
            key,val=i.items()[0]
            print "<arg type=\"%s\" name=\"%s\" />"%(Translate(key),Translate(val))
                    
        print endstr
        
        
    def __repr__(self):
        s=""
        for i in self.Args:
            key,val=i.items()[0]
            s=s+key+" "+val+","
        s=s[:-1]
        val="[%s %s (%s); virt=%d method=%d]"%(self.Rtype,self.Name,s,self.IsVirtual,self.IsMethod)
        return val;
    
class Ctor:
    """ Create a function desc, using Function(name,rtype)
    Add arguments in the <obj>.Args.append() """    
    def __init__(self,name):
        self.Name=name;
        self.Args=[]#list of dicts

    def emit_xml(self):
        print "<ctor  name=\"%s\"  >"%(self.Name)
        
        #TODO: Add support for Default arguments.
        for i in self.Args:
            key,val=i.items()[0]
            print "<arg type=\"%s\" name=\"%s\" />"%(key,val)
        print "</ctor>"


    def __repr__(self):
        s=""
        for i in self.Args:
            key,val=i.items()[0]
            s=s+key+" "+val+","
        s=s[:-1]
        val="[%s (  %s )]"%(self.Name,s);
        return val;

class Enum:
    """ Create a function desc, using Function(name,rtype)
    Add arguments in the <obj>.Args.append() """
    count=0;
    def  __init__(self,name):
        Enum.count=Enum.count+1
        if (name==""):
            name=name+str(Enum.count)
            
        self.Name=name;
        self.Elem=[]        
        

    def emit_xml(self):
        print "<enum  name=\"%s\" >"%(self.Name)

        #TODO: Add support for Default arguments.
        for i in self.Elem:
            key,val=i.items()[0]
            print "<enumerator  name=\"%s\"  value=\"%s\"/>"%(key,val)
            
        print "</enum>"

        
    def __repr__(self):
        s=""
        for i in self.Elem:
            key,val=i.items()[0]
            s=s+key+" "+val+","
        s=s[:-1]
        val="[%s { %s }]"%(self.Name,s);
        return val
    
        
        
class FLTKDescriptor:
    count=0
    def __init__(self,namespace,classname,parentname):
        if FLTKDescriptor.count==0:
            print "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
            FLTKDescriptor.count=FLTKDescriptor.count+1;
        self.IsClass=False;
        
        #these 3 hold, crude representations of things.
        self.PublicInstanceMethods=[]
        self.PublicStaticMethods=[]
        self.Enumerators=[]
        
        #these 4 hold fine representation of things.
        self.Ctors=[] #Ctor objects 
        self.PublicMethods=[]#Function Object
        self.StaticMethods=[]#Function Object
        self.Enums=[]#Enum Object
        
        if(namespace=="(null)"):
            #self.NameSpace="-----None-------";
            self.NameSpace="";
        else:
            self.NameSpace=namespace.split(" ")[1].strip();
            
        if(classname=="(null)"):
            #self.ClassName="-----None-------";
            self.ClassName="";
        else:
            self.ClassName=classname.split(":")[0].strip();
            self.IsClass=True;
            
        if(parentname=="(null)"):
            #self.ParentName="-----None-------";
            self.ParentName="";
        else:
            self.ParentName=parentname.replace(": public ","").split(" ")[0]
        
    def __repr__(self):
        val="Namespace [%s] Class [%s] Parent [%s]"%(self.NameSpace,
                                                       self.ClassName,
                                                       self.ParentName);
        return "<!-- "+val+"-->";
        
    def generate(self):
        self.parse_ctors();
        self.parse_functions();
        self.parse_enums();

        #"""Generate XML for other bindings to use."""
        #self.generate_xml();

        #"""generate FLTK2-C code"""
        #self.generate_c_code();

        
        #"""generate Octave-FLTK2 code"""
        self.generate_octave_code();
        
        return

    def generate_xml(self):
        """<?xml version="1.0" encoding="UTF-8"?>"""        
        print "<FLTK api-version=\"2.0\">"
        if (self.IsClass):
            print "<class name=\"%s\" parent=\"%s\" namespace=\"%s\">"%(self.ClassName,self.ParentName,self.NameSpace);

        for ctor in self.Ctors:
            ctor.emit_xml();
            
        for method in self.PublicMethods:
            method.emit_xml();

        for method in self.StaticMethods:
            method.emit_xml();
            
        for enum in self.Enums:
            enum.emit_xml();

            
        if (self.IsClass):
            print "</class>"
            
        print "</FLTK>"

        return
        
    def parse_ctors(self):
        #mark Ctors for deletion from Public Instance Methods.        
        deletion=[]; #and deletion from Public Static Methods

        if not self.IsClass:
            return
        
        for i in range(0,len(self.PublicInstanceMethods)):
            func=self.PublicInstanceMethods[i].strip()
            parts=func.split("(");

            p0=parts[0].strip();
            p0=p0.split(" ");
            if len(p0) > 1:
                continue
            p0=p0[0]
            if p0.find(self.ClassName) < 0 :
                continue

            #remove the ctor definitions from the list.
            deletion.append(i);
            
            name=p0;
            args=parts[1].split(")")[0].split(",");

            ctor=Ctor(name);
            count=0;
            for z in args:
                count=count+1;
                z=z.strip();                
                if (z.rfind("*") >= 0):
                    formals=[z[:z.rfind("*")+1], z[z.rfind("*")+1:]];
                elif (z.rfind("&")>=0):
                    formals=[z[:z.rfind("&")+1], z[z.rfind("&")+1:]];
                else:
                    formals=z.split(" ");
                    
                if formals[0]!="void" and formals[0]!="":
                    formals.append("_"+formals[0]+str(count));
                else:
                    formals.append("");
                    
                ctor.Args.append({formals[0].strip():formals[1].strip()})
            
            self.Ctors.append(ctor);
            #print ctor;


        count=0;
        for z in deletion:
            del self.PublicInstanceMethods[z-count];
            count=count+1
        return


    def parse_functions(self):
        # Look for Virtual, Pure Virtual functions.
        # 
        for i in range(0,len(self.PublicInstanceMethods)):
            func=self.PublicInstanceMethods[i].strip()
            parts=func.split("(")
            
            name=parts[0].strip()
            rtype=name[:name.rfind(" ")]
            name=name[name.rfind(" "):].strip()            
            args=parts[1].split(")")[0].split(",");


            if(rtype.find("inline")>=0):
                rtype=rtype.replace("inline","");
                
            #Function is an instance method.
            function=Function(name,rtype,True);
            if(rtype.find("virtual")>=0):
                function.IsVirtual=True;

            count=0;
            for i in args:
                i=i.strip();
                count=count+1;
                if (i.rfind("*") >= 0):
                    formals=[i[:i.rfind("*")+1], i[i.rfind("*")+1:]];
                elif (i.rfind("&")>=0):
                    formals=[i[:i.rfind("&")+1], i[i.rfind("&")+1:]];
                else:
                    formals=i.split(" ");
                if(len(formals)==1): #meaning, no formal args exist..
                    if formals[0]!="void" and formals[0]!="":
                        formals.append("_"+formals[0]+str(count));
                    else:
                        formals.append("")
                    
                function.Args.append({formals[0].strip():formals[1].strip()})
                
            self.PublicMethods.append(function);
            #print function;

        
        for i in range(0,len(self.PublicStaticMethods)):
            func=self.PublicStaticMethods[i].strip()
            parts=func.split("(")
            
            name=parts[0]
            rtype=name[:name.rfind(" ")].replace("static","")
            name=name[name.rfind(" "):]
            
            args=parts[1].split(")")[0].split(",");

            #this function is *NOT* an instance method.
            function=Function(name,rtype,False);
            
            count=0;
            for i in args:
                count=count+1;
                i=i.strip();
                if (i.rfind("*") >= 0):
                    formals=[i[:i.rfind("*")+1], i[i.rfind("*")+1:]];
                elif (i.rfind("&")>=0):
                    formals=[i[:i.rfind("&")+1], i[i.rfind("&")+1:]];
                else:
                    formals=i.split(" ");
                if(len(formals)==1): #meaning, no formal args exist..
                    if formals[0]!="void" and formals[0]!="":
                        formals.append("_"+formals[0]+str(count));
                    else:
                        formals.append("");
                function.Args.append({formals[0].strip():formals[1].strip()})
                
            self.StaticMethods.append(function);
            #print function;

        return


        self.PublicMethods=[]
        self.StaticMethods=[]
        self.Enums=[]

    def parse_enums(self):
        
        for i in range(0,len(self.Enumerators)):
            if self.IsClass:
                enum=Enum(self.NameSpace+"."+self.ClassName);
            else:
                enum=Enum("");

            ecount=0;
            val=self.Enumerators[i];
            for e in val:
                parts=e[0].split("=");
                if(len(parts)==1):
                    parts.append(str(ecount));
                else:
                    parts[1]=parts[1].replace(",","");
                
                enum.Elem.append({parts[0].strip():parts[1].strip()})
                
            self.Enums.append(enum);
            #print enum;
            
        return

    def generate_octave_code(self):
        """generate C++ LibOctave for Octave-FLTK2"""
        
        
        
        return

    
    def generate_octave_code(self):
        """generate C++ LibOctave for Octave-FLTK2"""
        
        for method in self.PublicMethods:
            method.emit_octave(self.ClassName);

        for method in self.StaticMethods:
            method.emit_octave(self.ClassName);
            
        for enum in self.Enums:
            enum.emit_octave(self.ClassName);
            
        if (self.IsClass):
            print "DEFUN_DLD(%s,args,,\"Loading Class %s\"){"%(self.ClassName,self.ClassName);
        else:
            print "DEFUN_DLD( @MODULE@ ,args,,\"Loading Module %s\"){"

            #must do arg-check & if args match, create an Object of this type,
            #and throw back to the user.
            for ctor in self.Ctors:
                ctor.emit_octave(self.ClassName);


            for method in self.PublicMethods:
                print "octave_dynamic_loader::load(\"%s\","");"%(method.Name);

            for method in self.StaticMethods:
                print "octave_dynamic_loader::load(\"%s\","");"%(method.Name);
            
            for enum in self.Enums:
                print "/* Skipping Enum [ %s ]*/"%(enum.Name);
            
            
        print "}"

        return
        
