%{

#if 0
  Lexer Rules: Longest Match, Precendence of Rules.
#endif

#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<glib.h>
#include<assert.h>
#include<ctype.h>

struct StateStruct
{
  int public_flag;
  int private_protected_flag;
  int virtual_flag;
  int in_class;
  int static_flag;
  int in_enum;
  char *this_class;
  char *parent_class;
  char *namespc;
  GSList *funclist; /* public-static , public-inst, public-constructor */
  GSList *enumlist; /* List of [public enumeration] Lists */
  GSList *currenumlist;
};

struct StateStruct LexerState={0,};

void namespace_handler(void)
  {
    if(LexerState.namespc)
      free(LexerState.namespc);
    LexerState.namespc=strdup(yytext);
  }

void enum_handler(const char *state)
  {
    static char *name=NULL;

    
    fprintf(stderr,"Enum called in %s \n",state);
    if(!strcmp(state,"add"))
      {
	assert(LexerState.currenumlist!=NULL);
	fprintf(stderr,"Enum [%s]\n",yytext);
	g_slist_append(LexerState.currenumlist,(void *)strdup(yytext));
      }
    else if(!strcmp(state,"begin"))
      {
	if(name)
	  free(name);
	name=strdup(yytext);	  
	fprintf(stderr,"Starting Enum[ %s ]\n",name);

	if(LexerState.currenumlist==NULL)
	  LexerState.currenumlist=g_slist_alloc();
      }
    else if(!strcmp(state,"end"))
      {
	fprintf(stderr,"Closing Enum[ %s ]\n",yytext);
	if(name)
	  free(name);
	g_slist_append(LexerState.enumlist,
		       (void *)LexerState.currenumlist);
	LexerState.currenumlist=NULL;
	name=NULL;
      }
  }

void class_handler(void)
{
  char *cname=NULL;
  cname=strstr(yytext,"class");
  cname=cname+6;
  LexerState.this_class=(char *)strdup(cname);

  cname=strstr(yytext,": ");
  if(cname)
    cname=strdup(cname);
  LexerState.parent_class=cname;
  fprintf(stderr,"Class data[ %s ]\n",yytext);
}

void function_handler(void)
{
  char *ptr=yytext;
  char *xptr=NULL;
  for(;*ptr!='\0';((*ptr=='\n'||*ptr=='\t')?*ptr=' ':*ptr,ptr++));	    		    
  /* Removing 'extern' kw from the list */
  xptr=strstr(yytext,"extern");
  if(xptr)	
    yytext=xptr+sizeof("extern");

  g_slist_append(LexerState.funclist,(char *)strdup(yytext));
  fprintf(stderr,"%s\n",yytext);
}


/* Single point entry to produce output */
void pythonize(void)
{
  static int count=0;
  GSList *iter=NULL;
/* Dump the list, name space, functions etc in python code.*/
printf("#! /usr/bin/python \n");
printf("from FLTKDescriptor import *\n");
printf("widget%d =FLTKDescriptor(\"%s\",\"%s\",\"%s\");\n",
       count,
       LexerState.namespc,
       LexerState.this_class,
       LexerState.parent_class);

 if(LexerState.funclist!=NULL &&
    LexerState.funclist->next!=NULL && 
	LexerState.funclist->next->data!=NULL)
   {
     for(iter=LexerState.funclist->next;iter!=NULL;iter=iter->next)
       {
	 if(strstr(iter->data,"static ")!=NULL)
	   {
	     printf("widget%d.PublicStaticMethods.append(\"%s\");\n",
		    count,
		    (char *)iter->data);
	   }
	 else
	   {
	     printf("widget%d.PublicInstanceMethods.append(\"%s\");\n",
		    count,
		    (char *)iter->data);
	   }
       }
   }

 if(LexerState.enumlist!=NULL &&
    LexerState.enumlist->next!=NULL &&
        LexerState.enumlist->next->data!=NULL)
   {
     for (iter=LexerState.enumlist->next;iter!=NULL;iter=iter->next)
       {
	 GSList *iter2=NULL;
	 GSList *enumlist=(void *)iter->data;
	 
	 if(enumlist==NULL)
	    break;

	 printf("widget%d.Enumerators.append([\n",count);
	 for(iter2=enumlist->next;iter2!=NULL;iter2=iter2->next)
	   {
	     printf("\t\t[\"%s\"],\n",(char *)iter2->data);
	   }
	 printf("\t\t]);\n");
       }
   }
 printf("print widget%d;\n",count);
 printf("widget%d.generate();\n",count);
 count++; 
}

/* Main Code Generation Module */
void generate_code(void)
{
  GSList *iter=NULL;
  pythonize();
  //xmlize();
  //SWIGize();
  
   for (iter=LexerState.enumlist->next;iter!=NULL;iter=iter->next)
   {
     GSList *iter2=NULL;
     GSList *enumlist=(void *)iter->data;
     
     for(iter2=enumlist->next;iter2!=NULL;iter2=iter2->next)
	 free(iter2->data);
   }

   g_slist_free(LexerState.enumlist);
   g_slist_free(LexerState.funclist);
   
   LexerState.enumlist=g_slist_alloc();
   LexerState.funclist=g_slist_alloc();
   return;
}


%}

%%
^[ \t\n]*[}][;]* { if (LexerState.in_enum) 
                   {
                    LexerState.in_enum=0;
                    enum_handler("end");
                    fprintf(stderr,"closing enumerator\n");
                    }
                else if(LexerState.in_class)
                    {
                     fprintf(stderr,"Class ends\n"); 
                     LexerState.in_class=0;
                    }
                   //fprintf(stderr,"Class ends [?] with %s\n",yytext); 
                 }
^class[A-Za-z\ :]+[{]  {                            
                         fprintf(stderr,"Class begins: %s\n",yytext);
                         LexerState.in_class=1;
 
                        //support for multiple classes in one header.
                        if(LexerState.this_class!=NULL)
                        {
                           fprintf(stderr,
			     "Completing code generation for %s class\n",
			      LexerState.this_class);
                           generate_code();
                        }
                        class_handler();
                       }

^typedef.*  { /* Ignore Type Defs ATM */ 
               fprintf(stderr,"/* Ignore Type Defs ATM */ \n");
            }

^public:.*        { LexerState.public_flag=1;
                    LexerState.private_protected_flag=0;
		    fprintf(stderr,"Handled Public: set flag\n");
                   }
^private:.*        { LexerState.private_protected_flag=1;
                     LexerState.public_flag=0;
		     fprintf(stderr,"Handled private: set flag\n");
                   }
^protected:.*      { LexerState.private_protected_flag=1;
                    fprintf(stderr,"Handled protected: set flag\n"); 
                   }
^namespace.*      { LexerState.namespc=strdup(yytext);
                    fprintf(stderr,"Handled Namespace= [ %s ] \n",yytext);
                    namespace_handler();
                  } 
^[A-Za-z_, ]+[=]*[A-Za-z_,0-9|\ ]* {
                                   if(LexerState.in_enum)
                                    {
                                      enum_handler("add");
                                    }
                                }
^[ \t\n]+enum.*[{]  {
                      if(!LexerState.private_protected_flag)
                          {
                            LexerState.in_enum=1;
                            enum_handler("begin");
                          }
                     }
^[\ a-zA-Z0-9\_][a-zA-Z0-9\ \*\_]*[(][a-zA-Z0-9\_\ =&\t,\*\n\.]*[)][\ a-zA-Z0-9\_]*               {                          
                          if(LexerState.private_protected_flag)
			    {
			      fprintf(stderr,"PRIVATE: [ %s ]\n",yytext);
			    }
                           function_handler();
                        }
^#.* {
      fprintf(stderr,"#ing: %s\n",yytext);
     }
"\n" { /* skipping */ }
[ ]+ {  /* skipped */}
. { /* skipped  */ 

    //fprintf(stderr,"Skipping: %s\n",yytext);
  } 

%%



int main(int argc,char *argv[])
{
  extern struct StateStruct LexerState;
  extern FILE * yyin;

 if(argc==2)
 { 
      yyin=fopen(argv[1],"r");     
 }


  if(!LexerState.funclist)
    LexerState.funclist=g_slist_alloc();

  if(!LexerState.enumlist)
    LexerState.enumlist=g_slist_alloc();

  while(!feof(yyin))
      yylex();
  generate_code();
  return 0;
}

